import requests
import json
from typing import TypeVar, Generic, Optional, Dict
from collections import namedtuple
from ..my_pydantic import BaseModel


T = TypeVar('T')

class WaiiHttpClient(Generic[T]):
    instance = None

    def __init__(self, url: str, apiKey: str):
        WaiiHttpClient.instance = self
        self.url = url
        self.apiKey = apiKey
        self.timeout = 150000000
        self.scope = ''
        self.orgId = ''
        self.userId = ''
        self.impersonateUserId = ''

    @classmethod
    def get_instance(cls, url: str = None, apiKey: str = None):
        if cls.instance is None or (url and url != cls.instance.url) or (apiKey and apiKey != cls.instance.apiKey):
            cls.instance = WaiiHttpClient(url, apiKey)
        return cls.instance

    def set_scope(self, scope: str):
        self.scope = scope

    def get_scope(self):
        return self.scope

    def set_org_id(self, orgId: str):
        self.orgId = orgId

    def set_user_id(self, userId: str):
        self.userId = userId

    def set_impersonate_user_id(self, userId: str):
        self.impersonateUserId = userId

    def common_fetch(
            self, 
            endpoint: str,
            params: Dict,
            cls: BaseModel = None,
            need_scope: bool = True,
            ret_json: bool = False
        ) -> Optional[T]:

        if need_scope:
            if not self.scope or self.scope.strip() == '':
                raise Exception("You need to activate connection first, use `WAII.Database.activate_connection(...)`")
            params['scope'] = self.scope
        params['org_id'] = self.orgId
        params['user_id'] = self.userId

        headers = {'Content-Type': 'application/json'}
        if self.apiKey:
            headers['Authorization'] = f'Bearer {self.apiKey}'
        if self.impersonateUserId:
            headers['x-waii-impersonate-user'] = self.impersonateUserId
        response = requests.post(self.url + endpoint, headers=headers, data=json.dumps(params, default=vars), timeout=self.timeout/1000)  # timeout is in seconds

        if response.status_code != 200:
            try:
                print(response)
                error = response.json()
                raise Exception(error.get('detail', ''))
            except json.JSONDecodeError:
                raise Exception(response.text)
        try:
            if cls:
                result: T = cls(**response.json())
            else:
                if not ret_json:
                    result: T = json.loads(response.text, object_hook=lambda d: namedtuple('X', d.keys())(*d.values()))
                else:
                    result: T = response.json()
            return result
        except json.JSONDecodeError:
            raise Exception("Invalid response received.")



from enum import Enum
from typing import Optional, List

from ..my_pydantic import BaseModel

from waii_sdk_py.common import CommonRequest, CommonResponse
from waii_sdk_py.database import TableName
from waii_sdk_py.waii_http_client import WaiiHttpClient

UPDATE_TABLE_ACCESS_RULES_ENDPOINT = "update-table-access-rules"
REMOVE_TABLE_ACCESS_RULES_ENDPOINT = "remove-table-access-rules"
LIST_TABLE_ACCESS_RULES_ENDPOINT = "list-table-access-rules"


class TableAccessRuleType(str, Enum):
    filter = "filter"  # protect all access with a filter
    block = "block"  # stop all access from the identified users


class TableAccessRule(BaseModel):
    id: Optional[str]
    name: str
    table: TableName
    org_id: str = '*'
    tenant_id: str = '*'
    user_id: str = '*'
    type: TableAccessRuleType
    expression: Optional[str]


class UpdateTableAccessRuleRequest(CommonRequest):
    rules: List[TableAccessRule]


class RemoveTableAccessRuleRequest(CommonRequest):
    rules: List[str]


class ListTableAccessRuleRequest(CommonRequest):
    table: Optional[TableName]
    ids: Optional[List[str]]


class ListTableAccessRuleResponse(CommonResponse):
    rules: Optional[List[TableAccessRule]]


class AccessRuleImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def update_table_access_rules(
            self, params: UpdateTableAccessRuleRequest
    ) -> CommonResponse:
        return self.http_client.common_fetch(
            UPDATE_TABLE_ACCESS_RULES_ENDPOINT, params.__dict__, CommonResponse
        )

    def remove_table_access_rules(
            self, params: RemoveTableAccessRuleRequest
    ) -> CommonResponse:
        return self.http_client.common_fetch(
            REMOVE_TABLE_ACCESS_RULES_ENDPOINT, params.__dict__, CommonResponse
        )

    def list_table_access_rules(
            self, params: ListTableAccessRuleRequest
    ) -> ListTableAccessRuleResponse:
        return self.http_client.common_fetch(
            LIST_TABLE_ACCESS_RULES_ENDPOINT, params.__dict__, ListTableAccessRuleResponse
        )

AccessRules = AccessRuleImpl(WaiiHttpClient.get_instance())



import requests

base_url = "http://localhost:9859"

# API endpoint URL
modify_db_conn_endpoint = f"{base_url}/api/update-db-connect-info"

update_table_endpoint = f"{base_url}/api/update-table-definitions"
generate_query_endpoint = f"{base_url}/api/generate-query"


def add_db_conn():
    # Data payload
    add_conn_data = {
        "updated": [
            {
                "account_name": "",
                "warehouse": "",
                "database": "push_test",
                "host": "test_host",
                "db_type": "postgresql",
                "push": True,
            }
        ],
        "validate_before_save": True,
    }
    # Adding Connection
    add_conn_response = requests.post(modify_db_conn_endpoint, json=add_conn_data)

    print("Adding db conn response")
    return add_conn_response


def remove_conn():
    # Removing connection
    remove_conn_data = {"removed": ["postgresql://push/test_host/push_test"]}
    print("remove connection")
    remove_conn_response = requests.post(modify_db_conn_endpoint, json=remove_conn_data)
    return remove_conn_response


def add_table():
    # Updating tables
    table_update_data = {
        "updated_tables": [
            {
                "name": {
                    "table_name": "Album",
                    "schema_name": "PUBLIC",
                    "database_name": "push_test",
                },
                "columns": [
                    {
                        "name": "AlbumId",
                        "type": "integer",
                        "comment": None,
                        "sample_values": None,
                        "description": "Unique identifier for the album.",
                        "description_update_source": "generated",
                    },
                    {
                        "name": "Title",
                        "type": "character varying",
                        "comment": None,
                        "sample_values": None,
                        "description": "Title of the album.",
                        "description_update_source": "generated",
                    },
                    {
                        "name": "ArtistId",
                        "type": "integer",
                        "comment": None,
                        "sample_values": None,
                        "description": "Identifier for the artist of the album.",
                        "description_update_source": "generated",
                    },
                ],
                "comment": None,
                "last_altered_time": 0,
                "refs": None,
                "constraints": [
                    {
                        "source": "database",
                        "table": {
                            "table_name": "Artist",
                            "schema_name": "PUBLIC",
                            "database_name": "push_test",
                        },
                        "cols": ["ArtistId"],
                        "constraint_type": None,
                        "src_table": {
                            "table_name": "Album",
                            "schema_name": "PUBLIC",
                            "database_name": "push_test",
                        },
                        "src_cols": ["ArtistId"],
                        "comment": None,
                    }
                ],
                "inferred_refs": [],
                "inferred_constraints": [],
                "description": 'The "Album" table contains information about music albums, including the album ID, artist ID, and title. Users can use this table to track and organize albums by various artists, making it easier to manage and access music collections.',
            }
        ],
        "scope": "postgresql://push/test_host/push_test",
    }
    updating_table_response = requests.post(
        update_table_endpoint, json=table_update_data
    )
    # add table definition response
    print("Updating table def Response")
    return updating_table_response


def gen_query():
    gen_query_data = {
        "ask": "give me 5 albums name",
        "search_context": [],
        "tweak_history": [],
        "parent_uuid": None,
        "model": "Automatic",
        "scope": "postgresql://push/test_host/push_test",
        "org_id": "",
        "user_id": "",
    }

    gen_query_response = requests.post(generate_query_endpoint, json=gen_query_data)

    print("Generating query response")
    return gen_query_response


def main():
    add_db_resp = add_db_conn()
    print(add_db_resp.status_code)

    remove_db_resp = remove_conn()
    print(remove_db_resp.status_code)
    add_db_resp = add_db_conn()
    print(add_db_resp.status_code)
    update_table_resp = add_table()
    print(update_table_resp.status_code)
    gen_query_resp = gen_query()
    print(gen_query_resp.status_code)
    print(gen_query_resp.json())


if __name__ == "__main__":
    main()



from enum import Enum
from typing import Optional, List

from ..my_pydantic import BaseModel

from ..common import LLMBasedRequest
from ..query import GetQueryResultResponse, GeneratedQuery
from ..database import CatalogDefinition, SearchContext
from ..semantic_context import GetSemanticContextResponse
from ..chart import ChartGenerationResponse, ChartType
from ..waii_http_client import WaiiHttpClient

CHAT_MESSAGE_ENDPOINT = "chat-message"

class ChatModule(str, Enum):
    DATA = "data"
    TABLES = "tables"
    QUERY = "query"
    PLOT = "plot"

class ChatRequest(LLMBasedRequest):
    ask: str

    # should we streaming the output?
    # no need to support for the first implementation
    streaming: bool = False

    # link to previous conversation, pick up where conversation left off
    parent_uuid: Optional[str]

    # optional chart type, default to plotly
    chart_type: Optional[ChartType]

    search_context: Optional[List[SearchContext]]

    modules : Optional[List[ChatModule]]


class ChatResponseData(BaseModel):
    data: Optional[GetQueryResultResponse]
    sql: Optional[GeneratedQuery]
    chart_spec: Optional[ChartGenerationResponse]
    semantic_context: Optional[GetSemanticContextResponse]
    catalog: Optional[CatalogDefinition]


class ChatResponse(BaseModel):
    # template response
    response: str
    response_data: Optional[ChatResponseData]
    is_new: Optional[bool] = False
    timestamp: int
    timestamp_ms: Optional[int]
    chat_uuid: str


class ChatImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def chat_message(self, params: ChatRequest) -> ChatResponse:
        return self.http_client.common_fetch(CHAT_MESSAGE_ENDPOINT, params.__dict__, ChatResponse)


Chat = ChatImpl(WaiiHttpClient.get_instance())


import functools
import math
import threading
import time
import traceback
from typing import Optional, List, Dict, Any, Union, Literal
from enum import Enum, IntEnum

from ..my_pydantic import BaseModel, Field

from ..common import CommonRequest, LLMBasedRequest
from ..database import SearchContext, TableName, ColumnDefinition, SchemaName
from ..semantic_context import SemanticStatement
from ..waii_http_client import WaiiHttpClient

GENERATE_ENDPOINT = "generate-query"
RUN_ENDPOINT = "run-query"
SUBMIT_ENDPOINT = "submit-query"
FAVORITE_ENDPOINT = "like-query"
DESCRIBE_ENDPOINT = "describe-query"
DIFF_ENDPOINT = "diff-query"
RESULTS_ENDPOINT = "get-query-result"
CANCEL_ENDPOINT = "cancel-query"
AUTOCOMPLETE_ENDPOINT = "auto-complete"
PERF_ENDPOINT = "get-query-performance"
TRANSCODE_ENDPOINT = "transcode-query"
PLOT_ENDPOINT = "python-plot"
GENERATE_QUESTION_ENDPOINT = "generate-questions"
GET_SIMILAR_QUERY_ENDPOINT = "get-similar-query"
RUN_QUERY_COMPILER_ENDPOINT = "run-query-compiler"
SEMANTIC_CONTEXT_CHECKER_ENDPOINT = "semantic-context-checker"


class DebugInfoType(str, Enum):
    learned_template = "learned_template"
    retry_info = "retry_info"
    equivalent = "equivalent"
    fixit_info = "fixit_info"
    query_gen_source = "query_gen_source"
    query_gen_model = "query_gen_model"
    empty_table_selection = "empty_table_selection"
    after_tweak_history = "after_considering_tweak_history"
    after_info_schema_check = "after_info_schema_check"
    after_embedding_match = "after_embedding_match"
    after_initial_table_selection = "after_initial_table_selection"
    after_iterative_table_selection = "after_iterative_table_selection"


class Tweak(BaseModel):
    sql: Optional[str] = None
    ask: Optional[str] = None


class TranscodeQueryRequest(LLMBasedRequest):
    search_context: Optional[List[SearchContext]] = None
    ask: Optional[str] = ""
    source_dialect: Optional[str] = None
    source_query: Optional[str] = None
    target_dialect: Optional[str] = None


class QueryGenerationRequest(LLMBasedRequest):
    search_context: Optional[List[SearchContext]] = None
    tweak_history: Optional[List[Tweak]] = None
    ask: Optional[str] = None
    uuid: Optional[str] = None
    dialect: Optional[str] = None
    parent_uuid: Optional[str] = None
    flags: Optional[Dict[str, Any]] = {}


class DescribeQueryRequest(CommonRequest):
    search_context: Optional[List[SearchContext]] = None
    current_schema: Optional[str] = None
    query: Optional[str] = None


class DescribeQueryResponse(BaseModel):
    summary: Optional[str] = None
    detailed_steps: Optional[List[str]] = None
    tables: Optional[List[TableName]] = None


class DiffQueryRequest(DescribeQueryRequest):
    search_context: Optional[List[SearchContext]] = None
    current_schema: Optional[str] = None
    query: Optional[str] = None
    previous_query: Optional[str] = None


class DiffQueryResponse(DescribeQueryResponse):
    summary: Optional[str] = None
    tables: Optional[List[TableName]] = None
    detailed_steps: Optional[List[str]] = None
    what_changed: Optional[str] = None


class CompilationError(BaseModel):
    message: str
    line: Optional[int] = None


class LLMUsageStatistics(BaseModel):
    # total tokens consumed by the LLM model
    token_total: Optional[int]


class Query(BaseModel):
    uuid: str
    ask: str
    query: str
    detailed_steps: Optional[List[str]]


class ConfidenceScore(BaseModel):
    log_prob_sum: Optional[float]
    token_count: Optional[int]
    confidence_value: Optional[float]

    def get_linear_probability(self):
        if self.token_count:
            avg_log_prob = self.log_prob_sum / self.token_count

            return math.exp(avg_log_prob)
        else:
            return 0.0


class AccessRuleProtectionState(str, Enum):
    protected = "protected"  # Query is completely protected regarding access rules
    unprotected = "unprotected"  # Query is not guaranteed to be protected, see error msg for details


class AccessRuleProtectionStatus(BaseModel):
    state: AccessRuleProtectionState
    msg: Optional[str]


class GeneratedQuery(BaseModel):
    uuid: Optional[str] = None
    liked: Optional[bool] = None
    tables: Optional[List[TableName]] = None
    semantic_context: Optional[List[SemanticStatement]] = None
    query: Optional[str] = None
    detailed_steps: Optional[List[str]] = None
    what_changed: Optional[str] = None
    compilation_errors: Optional[List[CompilationError]] = None
    is_new: Optional[bool] = None
    timestamp_ms: Optional[int] = None
    llm_usage_stats: Optional[LLMUsageStatistics] = None
    confidence_score: Optional[ConfidenceScore]
    debug_info: Optional[Dict[str, Any]] = {}
    http_client: Optional[Any] = Field(default=None, exclude=True)
    access_rule_protection_status: Optional[AccessRuleProtectionStatus]  # query protection status regarding access rules

    def run(self):
        return QueryImpl(self.http_client).run(RunQueryRequest(query=self.query))


class RunQueryRequest(CommonRequest):
    query: str
    session_id: Optional[str] = None
    current_schema: Optional[SchemaName] = None
    session_parameters: Optional[Dict[str, Any]] = None


class RunQueryResponse(BaseModel):
    query_id: Optional[str] = None


class GetQueryResultRequest(CommonRequest):
    query_id: str
    max_returned_rows: Optional[int] = 10000


class CancelQueryRequest(CommonRequest):
    query_id: str


class CancelQueryResponse(BaseModel):
    pass


class GetQueryResultResponse(BaseModel):
    rows: Optional[List[object]] = None
    more_rows: Optional[int] = None
    column_definitions: Optional[List[ColumnDefinition]] = None
    query_uuid: Optional[str] = None

    def to_pandas_df(self):
        import pandas as pd

        return pd.DataFrame(
            self.rows, columns=[col.name for col in self.column_definitions]
        )


class LikeQueryRequest(CommonRequest):
    # you need to specify either query_uuid or ask/query
    query_uuid: Optional[str]
    ask: Optional[str]
    query: Optional[str]

    liked: bool

    # do we want to rewrite the question before storing it? by default, it is True. If it is False, then we will store
    # the ask as-is
    rewrite_question: Optional[bool] = False
    detailed_steps: Optional[List[str]] = []


class LikeQueryResponse(BaseModel):
    pass


class AutoCompleteRequest(CommonRequest):
    text: str
    cursor_offset: Optional[int] = None
    dialect: Optional[str] = None
    search_context: Optional[List[SearchContext]] = None
    max_output_tokens: Optional[int] = None


class AutoCompleteResponse(BaseModel):
    text: Optional[str] = None


class QueryPerformanceRequest(CommonRequest):
    query_id: str


class QueryPerformanceResponse(BaseModel):
    summary: List[str]
    recommendations: List[str]
    query_text: str
    execution_time_ms: Optional[int]
    compilation_time_ms: Optional[int]


class PythonPlotRequest(LLMBasedRequest):
    ask: Optional[str]
    dataframe_rows: Optional[List[Dict[str, Any]]]
    dataframe_cols: Optional[List[ColumnDefinition]]


class PythonPlotResponse(BaseModel):
    # based on the request, return N plot script
    plots: Optional[List[str]]


class GeneratedQuestionComplexity(str, Enum):
    easy = "easy"
    medium = "medium"
    hard = "hard"


class GenerateQuestionRequest(CommonRequest):
    schema_name: str
    n_questions: Optional[int] = 10  # number of questions to generate
    complexity: Optional[GeneratedQuestionComplexity] = GeneratedQuestionComplexity.hard


class GeneratedQuestion(BaseModel):
    question: str
    complexity: GeneratedQuestionComplexity
    tables: Optional[List[TableName]]  # tables used in the question


class GenerateQuestionResponse(BaseModel):
    questions: Optional[List[GeneratedQuestion]]


class SimilarQueryResponse(BaseModel):
    qid: Optional[int]
    equivalent: Optional[bool]
    query: Optional[Query]


class CompilationStateFromDBEngine(IntEnum):
    UNKNOWN = 0,  # this happens when explain query itself failed (because of permission, etc.)
    COMPILABLE = 1,
    UNCOMPILABLE = 2


class CompilationErrorMsgFromDBEngine(BaseModel):
    state: CompilationStateFromDBEngine
    msg: Optional[str]


class RunQueryCompilerResponse(BaseModel):
    query: str
    errors: str
    should_compile: bool
    tables: Optional[List[TableName]]
    explain_error_msg: CompilationErrorMsgFromDBEngine


class SemanticContextCheckerRequest(LLMBasedRequest):
    ask: Optional[str]
    query: Optional[str]
    dialect: Optional[str]
    search_context: Optional[List[SearchContext]]
    flags: Optional[Dict[str, Any]]


def show_progress(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        verbose = kwargs.get("verbose", False)

        if verbose:
            stop_event = threading.Event()
            dot_thread = threading.Thread(target=_print_dot, args=(stop_event,))
            dot_thread.start()

        try:
            return func(*args, **kwargs)
        except Exception as e:
            print("\nFailed:", e)
            raise
        finally:
            if verbose:
                stop_event.set()
                dot_thread.join()

    return wrapper


def _print_dot(stop_event):
    while not stop_event.is_set():
        print(".", end="", flush=True)
        time.sleep(1)


class QueryImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    @show_progress
    def generate(self, params: QueryGenerationRequest, verbose=True) -> GeneratedQuery:
        generated = self.http_client.common_fetch(
            GENERATE_ENDPOINT, params.__dict__, GeneratedQuery
        )
        generated.http_client = self.http_client
        return generated

    @show_progress
    def run(self, params: RunQueryRequest, verbose=True) -> GetQueryResultResponse:
        return self.http_client.common_fetch(
            RUN_ENDPOINT, params.__dict__, GetQueryResultResponse
        )

    def like(self, params: LikeQueryRequest) -> LikeQueryResponse:
        return self.http_client.common_fetch(
            FAVORITE_ENDPOINT, params.__dict__, LikeQueryResponse
        )

    def submit(self, params: RunQueryRequest) -> RunQueryResponse:
        return self.http_client.common_fetch(
            SUBMIT_ENDPOINT, params.__dict__, RunQueryResponse
        )

    def get_results(self, params: GetQueryResultRequest) -> GetQueryResultResponse:
        return self.http_client.common_fetch(
            RESULTS_ENDPOINT, params.__dict__, GetQueryResultResponse
        )

    def cancel(self, params: CancelQueryRequest) -> CancelQueryResponse:
        return self.http_client.common_fetch(
            CANCEL_ENDPOINT, params.__dict__, CancelQueryResponse
        )

    def describe(self, params: DescribeQueryRequest) -> DescribeQueryResponse:
        return self.http_client.common_fetch(
            DESCRIBE_ENDPOINT, params.__dict__, DescribeQueryResponse
        )

    def auto_complete(self, params: AutoCompleteRequest) -> AutoCompleteResponse:
        return self.http_client.common_fetch(
            AUTOCOMPLETE_ENDPOINT, params.__dict__, AutoCompleteResponse
        )

    def diff(self, params: DiffQueryRequest) -> DiffQueryResponse:
        return self.http_client.common_fetch(
            DIFF_ENDPOINT, params.__dict__, DiffQueryResponse
        )

    def analyze_performance(
        self, params: QueryPerformanceRequest
    ) -> QueryPerformanceResponse:
        return self.http_client.common_fetch(
            PERF_ENDPOINT, params.__dict__, QueryPerformanceResponse
        )

    def transcode(self, params: TranscodeQueryRequest) -> GeneratedQuery:
        generated = self.http_client.common_fetch(
            TRANSCODE_ENDPOINT, params.__dict__, GeneratedQuery
        )
        generated.http_client = self.http_client
        return generated

    @show_progress
    def plot(
        self, df, ask=None, automatically_exec=True, verbose=True, max_retry=2, model=None
    ) -> str:
        if df is None or df.empty:
            raise ValueError("(Plot) Input dataframe is empty")

        # create ColumnDefinition from df.columns, use first row to get type
        cols = []
        for col in df.columns:
            cols.append(ColumnDefinition(name=col, type=df[col][0].__class__.__name__))

        params = PythonPlotRequest(dataframe_cols=cols, ask=ask, model=model)
        plot_response = self.http_client.common_fetch(
            PLOT_ENDPOINT, params.__dict__, PythonPlotResponse
        )
        p = plot_response.plots[0]

        # if the p include ``` ... ```, only include lines between them, handle the case like
        # ```python
        # ...
        # ```
        # or
        # ```
        # <python code>
        # ```

        if p.startswith("```"):
            p = p[3:]
            if p.startswith("python"):
                p = p[6:]
            p = p.strip()
            p = p[:-3]
            p = p.strip()

        retried = False
        if automatically_exec:
            try:
                exec(p)
            except Exception as e:
                if max_retry > 0:
                    retried = True
                    print(
                        f"Trying to fix error={str(e)}, Retry with max_retry={max_retry}"
                    )
                    fix_msg = "Fix the error and generate plot, find the following code and exception:"
                    if not ask:
                        ask = ""
                    return self.plot(
                        df,
                        ask=fix_msg
                        + f'"{ask}"'
                        + f"\ncode:```{p}```\nException:{str(e)}",
                        automatically_exec=automatically_exec,
                        verbose=verbose,
                        max_retry=max_retry - 1,
                    )
                else:
                    # print the code when not verbose (because finally will print the code if it is verbose)
                    if not verbose:
                        print("=== generated code ===")
                        print(p)
                    traceback.print_exc()
                    raise e

        if not retried and verbose:
            print("=== generated code ===")
            print(p)
        return p

    def generate_question(
        self, params: GenerateQuestionRequest
    ) -> GenerateQuestionResponse:
        return self.http_client.common_fetch(
            GENERATE_QUESTION_ENDPOINT, params.__dict__, GenerateQuestionResponse
        )

    def get_similar_query(
        self, params: QueryGenerationRequest
    ) -> SimilarQueryResponse:
        return self.http_client.common_fetch(
            GET_SIMILAR_QUERY_ENDPOINT, params.__dict__, SimilarQueryResponse
        )

    def run_query_compiler(
            self, params: RunQueryRequest
    ) -> RunQueryCompilerResponse:
        return self.http_client.common_fetch(
            RUN_QUERY_COMPILER_ENDPOINT, params.__dict__, RunQueryCompilerResponse
        )

    def handle_semantic_context_checker(
            self, params: SemanticContextCheckerRequest
    ) -> GeneratedQuery:
        return self.http_client.common_fetch(
            SEMANTIC_CONTEXT_CHECKER_ENDPOINT, params.__dict__, GeneratedQuery
        )


Query = QueryImpl(WaiiHttpClient.get_instance())


import unittest

from tests.common_test_utils import load_db_conn1, load_db_conn2, connect_db
from waii_sdk_py.database import Database
from waii_sdk_py.query import *
from waii_sdk_py.waii_sdk_py import Waii, WAII
"""
This class is to test https://doc.waii.ai/python/docs/multi-tenant-client-module
"""
class MultiTenantClientTest(unittest.TestCase):
    def setUp(self):
        self.movie_conn = load_db_conn1()
        self.chinook_conn = load_db_conn2()

        connect_db(self.movie_conn)
        connect_db(self.chinook_conn)

        movie_waii = Waii()
        chinook_waii = Waii()
        movie_waii.initialize(url="http://localhost:9859/api/")
        chinook_waii.initialize(url="http://localhost:9859/api/")
        result = movie_waii.database.get_connections()
        WAII.initialize(url="http://localhost:9859/api/")
        self.movie_waii = movie_waii
        self.chinook_waii = chinook_waii

    def test_movie_generate(self):
        self.movie_waii.database.activate_connection(self.movie_conn.key)
        params = QueryGenerationRequest(
            ask="Give me 5 movie names sorted by movie name"
        )
        result = self.movie_waii.query.generate(params)
        self.assertIsInstance(result, GeneratedQuery)
        assert result.uuid is not None
        assert len(result.detailed_steps) > 0
        assert len(result.query) > 0
        assert "sdk_test.public.movies" in result.query.lower()
        assert len(result.tables) > 0

        params = LikeQueryRequest(query_uuid=result.uuid, liked=True)
        result = self.movie_waii.query.like(params)
        self.assertIsInstance(result, LikeQueryResponse)

    def test_movie_run(self):
        self.movie_waii.database.activate_connection(self.movie_conn.key)
        params = QueryGenerationRequest(
            ask="Give me 5 movie names sorted by movie name"
        )
        result = self.movie_waii.query.generate(params)
        params = RunQueryRequest(query=result.query)
        result = self.movie_waii.query.run(params)
        self.assertIsInstance(result, GetQueryResultResponse)
        assert len(result.column_definitions) > 0
        assert "Incept" in str(result.rows[0])

    def test_chinook_generate(self):
        self.chinook_waii.database.activate_connection(self.chinook_conn.key)
        params = QueryGenerationRequest(ask="Give me 5 album names sorted by name")
        result = self.chinook_waii.query.generate(params)
        self.assertIsInstance(result, GeneratedQuery)
        assert result.uuid is not None
        assert len(result.detailed_steps) > 0
        assert len(result.query) > 0
        assert "public" in result.query.lower()
        assert len(result.tables) > 0

        params = LikeQueryRequest(query_uuid=result.uuid, liked=True)
        result = self.chinook_waii.query.like(params)
        self.assertIsInstance(result, LikeQueryResponse)

    def test_chinook_run(self):
        self.chinook_waii.database.activate_connection(self.chinook_conn.key)
        params = QueryGenerationRequest(ask="Give me 5 album names sorted by name")
        result = self.chinook_waii.query.generate(params)
        params = RunQueryRequest(query=result.query)
        result = self.chinook_waii.query.run(params)
        self.assertIsInstance(result, GetQueryResultResponse)
        assert len(result.column_definitions) > 0
        assert "...And" in str(result.rows[0])

    def test_legacy_generate(self):
        WAII.Database.activate_connection(self.movie_conn.key)
        params = QueryGenerationRequest(
            ask="Give me 5 movie names sorted by movie name"
        )
        result = WAII.Query.generate(params)
        self.assertIsInstance(result, GeneratedQuery)
        assert result.uuid is not None
        assert len(result.detailed_steps) > 0
        assert len(result.query) > 0
        assert "sdk_test." in result.query.lower()
        assert len(result.tables) > 0

        params = LikeQueryRequest(query_uuid=result.uuid, liked=True)
        result = WAII.Query.like(params)
        self.assertIsInstance(result, LikeQueryResponse)

    def test_legacy_run(self):
        WAII.Database.activate_connection(self.movie_conn.key)
        params = QueryGenerationRequest(
            ask="Give me 5 movie names sorted by movie name"
        )
        result = WAII.Query.generate(params)
        params = RunQueryRequest(query=result.query)
        result = WAII.Query.run(params)
        self.assertIsInstance(result, GetQueryResultResponse)
        assert len(result.column_definitions) > 0
        assert "Incepti" in str(result.rows[0])

    def test_legacy_run_without_WAII(self):
        Database.activate_connection(self.movie_conn.key)
        params = QueryGenerationRequest(
            ask="Give me 5 movie names sorted by movie name"
        )
        result = Query.generate(params)
        params = RunQueryRequest(query=result.query)
        result = Query.run(params)
        self.assertIsInstance(result, GetQueryResultResponse)
        assert len(result.column_definitions) > 0
        assert "Incep" in str(result.rows[0])

    def test_activate_connection(self):
        Database.activate_connection(self.movie_conn.key)
        result = Database.get_activated_connection()
        assert (self.movie_conn.key == result)



from enum import Enum
from typing import Optional, Literal, List, Union, Dict, Any

from ..database import ColumnDefinition
from ..my_pydantic import BaseModel

from ..common import LLMBasedRequest
from ..waii_http_client import WaiiHttpClient

GENERATE_CHART_ENDPOINT = "generate-chart"


class ChartType(str, Enum):
    METABASE = "metabase"
    SUPERSET = "superset"
    PLOTLY = "plotly"


class SuperSetChartSpec(BaseModel):
    spec_type: Literal['superset']
    plot_type: Optional[str]
    metrics: Optional[List[str]]
    dimensions: Optional[List[str]]
    chart_name: Optional[str]
    color_hex: Optional[str]
    x_axis: Optional[str]
    y_axis: Optional[str]
    grid_style: Optional[str]
    stacked: Optional[bool]
    width: Optional[int]
    height: Optional[int]


class MetabaseChartSpec(BaseModel):
    spec_type: Literal['metabase']
    plot_type: Optional[str]
    metric: Optional[str]
    dimension: Optional[str]
    name: Optional[str]
    color_hex: Optional[str]


class PlotlyChartSpec(BaseModel):
    spec_type: Literal['plotly'] = 'plotly'
    plot: Optional[str]


class ChartTweak(BaseModel):
    ask: Optional[str]
    chart_spec: Optional[Union[SuperSetChartSpec, MetabaseChartSpec, PlotlyChartSpec]]

    def __str__(self):
        return f"previous ask={self.ask}, previous chart_spec={self.chart_spec})\n"


class ChartGenerationRequest(LLMBasedRequest):
    sql: Optional[str]
    ask: Optional[str]
    dataframe_rows: Optional[List[Dict[str, Any]]]
    dataframe_cols: Optional[List[ColumnDefinition]]
    chart_type: Optional[ChartType]
    parent_uuid: Optional[str]
    tweak_history: Optional[List[ChartTweak]]


class ChartGenerationResponse(BaseModel):
    uuid: str
    timestamp: Optional[int]
    timestamp_ms: Optional[int]
    chart_spec: Optional[Union[SuperSetChartSpec, MetabaseChartSpec, PlotlyChartSpec]]


class ChartImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def generate_chart(
        self, df, ask=None, sql=None, chart_type=None, parent_uuid=None, tweak_history=None,
    ) -> ChartGenerationResponse:

        #Remove duplicate columns
        df = df.loc[:, ~df.columns.duplicated()]

        cols = []
        for col in df.columns:
            cols.append(ColumnDefinition(name=col, type=df[col][0].__class__.__name__))

        params = ChartGenerationRequest(dataframe_cols=cols,
                                        dataframe_rows=df.to_dict(orient='records'),
                                        ask=ask,
                                        chart_type=chart_type,
                                        parent_uuid=parent_uuid,
                                        tweak_history=tweak_history)

        params_dict = {k: v.value if isinstance(v, Enum) else v for k, v in params.dict().items() if v is not None}

        return self.http_client.common_fetch(
            GENERATE_CHART_ENDPOINT, params_dict, ChartGenerationResponse
        )


Chart = ChartImpl(WaiiHttpClient.get_instance())



import warnings

from waii_sdk_py.waii_http_client import WaiiHttpClient
from ..common import LLMBasedRequest, CommonRequest
from ..my_pydantic import BaseModel, PrivateAttr
import re
from typing import Optional, List, Dict, Any, Union
from urllib.parse import urlparse, parse_qs
from enum import Enum

from ..user import CommonResponse

MODIFY_DB_ENDPOINT = "update-db-connect-info"
GET_CATALOG_ENDPOINT = "get-table-definitions"
UPDATE_TABLE_DESCRIPTION_ENDPOINT = "update-table-description"
UPDATE_SCHEMA_DESCRIPTION_ENDPOINT = "update-schema-description"
UPDATE_COLUMN_DESCRIPTION_ENDPOINT = "update-column-description"
UPDATE_CONSTRAINT_ENDPOINT = "update-constraint"
UPDATE_TABLE_DEFINITION_ENDPOINT = "update-table-definitions"


class SchemaName(BaseModel):
    schema_name: str
    database_name: Optional[str]


class TableName(BaseModel):
    table_name: str
    schema_name: Optional[str]
    database_name: Optional[str]


class ColumnSampleValues(BaseModel):
    values: Optional[Dict[str, int]]


class ColumnDefinition(BaseModel):
    name: str
    type: str
    comment: Optional[str]
    description: Optional[str]
    sample_values: Optional[ColumnSampleValues]

    def __lt__(self, other):
        return self.name < other.name


class ConstraintType(str, Enum):
    primary = "primary"
    foreign = "foreign"


class ConstraintDetectorType(str, Enum):
    database = "database"
    inferred_llm = "inferred_llm"
    inferred_static = "inferred_static"
    inferred_query_history = "inferred_query_history"
    manual = "manual"


class TableReference(BaseModel):
    src_table: Optional[TableName]  # table name
    src_cols: Optional[List[str]]  # source table columns
    ref_table: Optional[TableName]  # ref table name
    ref_cols: Optional[List[str]]  # ref table columns
    source: Optional[ConstraintDetectorType]


class TableNameToDescription(BaseModel):
    name: str
    description: str


class SchemaDescription(BaseModel):
    summary: Optional[str]
    common_questions: Optional[List[str]]
    common_tables: Optional[List[TableNameToDescription]]


class Constraint(BaseModel):
    source: Optional[ConstraintDetectorType]
    table: Optional[TableName]
    cols: Optional[List[str]]
    constraint_type: Optional[ConstraintType]

    # for foreign key, it is the table and columns that the source table and cols
    src_table: Optional[TableName]  # table name
    src_cols: Optional[List[str]]  # source table columns

    # comment for the constraint
    comment: Optional[str]

    def __repr__(self):
        if self.constraint_type == ConstraintType.primary:
            return f"PK {self.table} ({self.cols} [{self.source.name}]"
        else:
            return f"FK {self.src_table} ({self.src_cols}) -> {self.table} ({self.cols}) [{self.source.name}]"



class TableDefinition(BaseModel):
    name: TableName
    columns: Optional[List[ColumnDefinition]]
    comment: Optional[str]
    last_altered_time: Optional[int]
    _refs: Optional[List[TableReference]] = PrivateAttr(default=[])
    constraints: Optional[List[Constraint]]
    inferred_refs: Optional[List[TableReference]]
    inferred_constraints: Optional[List[Constraint]]
    description: Optional[str]

    def __init__(self, **data):
        super().__init__(**data)
        refs_data = data.get("refs",None)
        refs = []
        if refs_data:
            for ref in refs_data:
                if type(ref) == TableReference:
                    refs.append(ref)
                else:
                    refs.append(TableReference(**ref))

        self._refs = refs

    @property
    def refs(self):
        warnings.warn(
            "The 'refs' attribute is deprecated and will be removed in a future release. Use 'constraints' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._refs

    @refs.setter
    def refs(self, value):
        warnings.warn(
            "The 'refs' attribute is deprecated and will be removed in a future release. Use 'constraints' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._refs = value

class SchemaDefinition(BaseModel):
    name: SchemaName
    tables: Optional[List[TableDefinition]]
    description: Optional[SchemaDescription]


class CatalogDefinition(BaseModel):
    name: str
    schemas: Optional[List[SchemaDefinition]]

class DBContentFilterScope(str, Enum):
    schema = "schema"
    table = "table"
    column = "column"

class DBContentFilterType(str, Enum):
    include = "include"
    exclude = "exclude"

class DBContentFilterActionType(str, Enum):
    # this means we will include/exclude the content
    visibility = "visibility"

    # this means we will do sample value or not
    sample_values = "sample_values"

class SearchContext(BaseModel):
    db_name: Optional[str] = "*"
    schema_name: Optional[str] = "*"
    table_name: Optional[str] = "*"



class DBContentFilter(BaseModel):
    filter_scope: DBContentFilterScope
    filter_type: DBContentFilterType = DBContentFilterType.include
    filter_action_type: DBContentFilterActionType = DBContentFilterActionType.visibility
    ignore_case: Optional[bool] = True
    pattern: str  # regex pattern
    search_context: Optional[List[SearchContext]]



class DBConnection(BaseModel):
    key: str
    db_type: str
    description: Optional[str] = None
    account_name: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    database: Optional[str] = None
    warehouse: Optional[str] = None
    role: Optional[str] = None
    path: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    parameters: Optional[Dict[str, Any]] = None
    sample_col_values: Optional[bool]
    push:Optional[bool] = False
    db_content_filters: Optional[List[DBContentFilter]]
    embedding_model: Optional[str]
    always_include_tables: Optional[List[TableName]]

class ModifyDBConnectionRequest(BaseModel):
    updated: Optional[List[DBConnection]] = None
    removed: Optional[List[str]] = None
    validate_before_save: Optional[bool] = None
    user_id: Optional[str] = None
    default_db_connection_key: Optional[str] = None


class SchemaIndexingStatus(BaseModel):
    n_pending_indexing_tables: int
    n_total_tables: int
    status: str


class DBConnectionIndexingStatus(BaseModel):
    status: Optional[str]
    schema_status: Optional[Dict[str, SchemaIndexingStatus]]


class ModifyDBConnectionResponse(BaseModel):
    connectors: Optional[List[DBConnection]]
    diagnostics: Optional[str]
    default_db_connection_key: Optional[str]
    connector_status: Optional[Dict[str, DBConnectionIndexingStatus]]




class GetCatalogRequest(LLMBasedRequest):
    search_context: Optional[List[SearchContext]]

    # ask is to perform semantic search, if it is provided, then we will use it to search the table definition
    # we will try to find the same tables which can answer the ask
    ask: Optional[str]

    internal: bool = False

class GetDBConnectionRequest(BaseModel):
    pass


class GetDBConnectionResponse(BaseModel):
    connectors: Optional[List[DBConnection]]
    diagnostics: Optional[str]
    default_db_connection_key: Optional[str]
    connector_status: Optional[Dict[str, DBConnectionIndexingStatus]]


class GetCatalogResponse(BaseModel):
    catalogs: Optional[List[CatalogDefinition]] = None
    debug_info: Optional[Dict[str, Any]]


class UpdateTableDescriptionRequest(BaseModel):
    table_name: TableName
    description: str


class UpdateTableDefinitionRequest(BaseModel):
    updated_tables: Optional[List[TableDefinition]]
    removed_tables: Optional[List[TableName]]


class ColumnDescription(BaseModel):
    column_name: str
    description: Optional[str]


class TableToColumnDescription(BaseModel):
    table_name: TableName
    column_descriptions: Optional[List[ColumnDescription]]


class UpdateColumnDescriptionRequest(BaseModel):
    col_descriptions: Optional[List[TableToColumnDescription]]


class UpdatedTableToCol(BaseModel):
    table_name: TableName
    column_names: Optional[List[str]]


class UpdateColumnDescriptionResponse(BaseModel):
    updated_table_to_cols: Optional[List[UpdatedTableToCol]]

class UpdateTableDefinitionResponse(BaseModel):
    updated_tables: Optional[List[TableName]]


class UpdateSchemaDescriptionRequest(BaseModel):
    schema_name: SchemaName
    description: str


class UpdateTableDescriptionResponse(BaseModel):
    pass


class UpdateSchemaDescriptionResponse(BaseModel):
    pass


class TableConstraints(BaseModel):
    table_name: TableName
    constraints: Optional[List[Constraint]]
    constraint_type: ConstraintType


class UpdateConstraintRequest(BaseModel):
    # updated constraints, it will replace the existing constraints
    updated_constraints: Optional[List[TableConstraints]]


class UpdateConstraintResponse(BaseModel):
    updated_tables: Optional[List[TableName]]


class RefreshDBConnectionRequest(BaseModel):
    db_conn_key: str


class DatabaseImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def modify_connections(
        self, params: ModifyDBConnectionRequest
    ) -> ModifyDBConnectionResponse:
        return self.http_client.common_fetch(
            MODIFY_DB_ENDPOINT, params.__dict__, ModifyDBConnectionResponse
        )

    def get_connections(
        self, params: GetDBConnectionRequest = GetDBConnectionRequest()
    ) -> GetDBConnectionResponse:
        return self.http_client.common_fetch(
            MODIFY_DB_ENDPOINT,
            params.__dict__,
            GetDBConnectionResponse,
            need_scope=False,
        )

    def activate_connection(self, key: str):
        self.http_client.set_scope(key)

    def get_activated_connection(self):
        return self.http_client.get_scope()

    def get_default_connection(self):
        return self.http_client.get_scope()

    def get_catalogs(
        self, params: GetCatalogRequest = GetCatalogRequest()
    ) -> GetCatalogResponse:
        return self.http_client.common_fetch(
            GET_CATALOG_ENDPOINT, params.__dict__, GetCatalogResponse
        )

    def update_table_description(
        self, params: UpdateTableDescriptionRequest
    ) -> UpdateTableDescriptionResponse:
        return self.http_client.common_fetch(
            UPDATE_TABLE_DESCRIPTION_ENDPOINT, params.__dict__, GetCatalogResponse
        )

    def update_table_definition(
        self, params:UpdateTableDefinitionRequest
    ) -> UpdateTableDefinitionResponse:
        return self.http_client.common_fetch(
            UPDATE_TABLE_DEFINITION_ENDPOINT, params.__dict__,UpdateTableDefinitionResponse
        )

    def update_schema_description(
        self, params: UpdateSchemaDescriptionRequest
    ) -> UpdateSchemaDescriptionResponse:
        return self.http_client.common_fetch(
            UPDATE_SCHEMA_DESCRIPTION_ENDPOINT, params.__dict__, GetCatalogResponse
        )

    def update_column_description(
        self, params: UpdateColumnDescriptionRequest
    ) -> UpdateColumnDescriptionResponse:
        return self.http_client.common_fetch(
            UPDATE_COLUMN_DESCRIPTION_ENDPOINT,
            params.__dict__,
            UpdateColumnDescriptionResponse,
        )

    def update_constraint(
        self, params: UpdateConstraintRequest
    ) -> UpdateConstraintResponse:
        return self.http_client.common_fetch(
            UPDATE_CONSTRAINT_ENDPOINT, params.__dict__, UpdateConstraintResponse
        )

    def refresh_db_connection(self):
        return self.http_client.common_fetch(
            "refresh-db-connection",
            {
                "db_conn_key": self.get_activated_connection(),
            },
            CommonResponse,
        )


Database = DatabaseImpl(WaiiHttpClient.get_instance())
